## This script implements the model used to evaluate the resource
## assignment policy described in M. Marzolla, S. Ferretti, G.
## D'Angelo. "Auction-Based Resource Allocation in Digital
## Ecosystems", Proc. 6th International Conference on MOBILe Wireless
## MiddleWARE, Operating Systems, and Applications (MobilWare 2013),
## pp. 20-27, nov 11-12 2013, Bologna, Italy, DOI http://dx.doi.org/10.1109/Mobilware.2013.16
##
## This script has been written in july-august 2011 by Moreno Marzolla
## <marzolla (at) cs.unibo.it> and released by the author in the Public
## Domain.
##
## This script calls ecosystem.mod using the GNU Linear Programming Kit (GLPK)

1; # not a function file

figure (1, "visible", "off"); # do not show plot window

page_screen_output(0); # avoid output pagination

##############################################################################
## Define an adjacency matrix M(i,j). The matrix is randomly defined
## (FIXME! use a scale free graph instead!), and is symmetric, so that
## if (i,j) can interact, so can (j,i). We set M(i,i) = 0, so that a
## user cannot sell resources to himself. Parameter density is the
## fraction of ones in the matrix (i.e., if density=0.2, then on average
## (M^2)*0.2 elements will be ones).
function M = make_overlay(N,density)
  M = discrete_rnd([0 1], [1-density density], [N N]);
  M = triu(M,1);
  M = M + M';
  assert( issymmetric(M) );
endfunction

##############################################################################
## Compute a bid bid(r,i). A bid represents a matching between requests
## and offers. A bid does not imply that all requests will be satisfied,
## e.g., because buyers want to buy more resource than the available
## quantity.
##
## INPUTS:
##
## Req(r,i)	amount of resource r required by i
## Off(r,j)	amount of resource r offered by j
## p(r,j)	unitary price of resource r offered by j
## reserve(i)	total reserve price of buyer i (i will not pay more than reserve(i) for ALL the resources he needs)
##
## OUTPUT:
##
## bid(r,i) = j iff user i requests Req(r,i) items of resource r from j
##		(if bid(r,i) == 0, then user i does not request any resource
##		from any seller)
function bid = make_bids( Req, reserve, Off, p, M )
  assert( size(Req) == size(Off) );
  assert( issquare(M) );
  N = columns(Req);
  R = rows(Req);
  buyers = find(any(Req>0,1)); # identify buyers
  bid = zeros(R,N);
  for i=buyers
    total_cost = 0; # amount of money spent by user i; must be <= reserve(i)
    ## loop over the resources
    for r=1:R
      ## Find neighbors which have at least Req(r,i) units of type r resource
      neighbors = find( (M(i,:)==1) & (Off(r,:) >= Req(r,i)));
      if ( length(neighbors) == 0 )
	total_cost = inf; # no type r resource available. Give up
	break;
      else
	prices = p(r,neighbors);
	[pmin imin] = min(prices);
	total_cost += pmin*Req(r,i);
	bid(r,i) = neighbors(imin);
      endif
    endfor
    if ( total_cost > reserve(i) )
      bid(:,i) = 0; # the total cost for user i is greater than the reserve price for i. Therefore, user i does not buy anything
    endif
  endfor
endfunction

##############################################################################
## Given a matching returned by find_matching, compute the total amount
## of resource of each type which is required to each node j
##
## OUTPUT:
##
## res(r,j)	total amount of resource r required to j
function res = compute_demands( bid, Req )
  R = rows(Req);
  N = columns(Req);
  res = zeros(R,N);
  for r=1:R
    for i=1:N   
      j = bid(r,i);
      if ( j>0 )
	res(r,j) += Req(r,i);
      endif
    endfor
  endfor
endfunction

##############################################################################
## Given a bid bid(r,i), resource requests Req(r,j) and selling prices
## p(r,j), compute res(i), the total amount of money spent by user i to
## buy the resources he needs (Req(1,i), ... Req(R,i)).
##
## INPUTS:
##
## bid(r,i) = j iff user i requests resource r from user j; if j=0, user i does not request resource r
## Req(r,i)	amount of type r resource requested by user i
## p(r,j)	unitary price of type r resource sold by user j
##
## OUTPUT:
##
## expenses(i)  total amount spent by buyer i
## incomes(j)	total amount earned by seller j
function [expenses incomes] = compute_expense( bid, Req, p )
  R = rows(Req);
  N = columns(Req);
  expenses = incomes = zeros(1,N);
  for r=1:R
    for i=1:N
      j = bid(r,i);
      if ( j>0 )
	expenses(i) += Req(r,i)*p(r,j);
	incomes(j) += Req(r,i)*p(r,j);
      endif
    endfor
  endfor
endfunction

##############################################################################
## Setup a new simulation round: create random offers and requests for R
## resource types shared by N users. The fraction of buyers is fb (i.e., if
## fb = 0.2, then N*0.2 users are buyers and N*0.8 are seller)
function [Off, Req] = setup_round( N, R, fb )
  ## randomly select N/10 "buyers" (i.e., nodes requesting resources)
  pp = randperm(N); # random permutation of 1..N
  buyers = pp(1:fix(N*fb));
  sellers = pp(fix(N*fb)+1:end);

  ## Define the amounts of resources offered
  Off = unidrnd(10,R,N);
  ## Define the amounts of resources required
  Req = unidrnd(10,R,N);
  ## Only buyers request something
  Req(:,sellers) = 0;
  ## Only sellers offer something
  Off(:,buyers) = 0;
endfunction

##############################################################################
## Check whether a bid is correct
function res = iscorrect( bid, Req, Off, M)
  res = 1;
  R = rows(Req);
  N = columns(Req);
  for i=1:N
    for r=1:R
      if ( (Req(r,i) > 0) && (bid(r,i) > 0) )
	j = bid(r,i);
	assert( M(i,j) == 1 );
	assert( Req(r,i) <= Off(r,j) );
	Off(r,j) -= Req(r,i);
      endif
    endfor
  endfor
endfunction

##############################################################################
## Dumps the content of two-dimensional matrix X(i,j) in a file
## identified by handle f. The output is in GNU MathProg format, and is
## associated to parameter "name".
function dump_glpk_matrix( f, name, X, tt )
  (nargin >= 3 && nargin <=4) || \
      print_usage();
  if ( nargin == 3 )
    tt = "int";
  endif
  fprintf(f, "param %s :", name);
  for l=1:columns(X)
    fprintf(f, " %d", l);
  endfor
  fprintf(f, " := \n");
  for k=1:rows(X)
    fprintf(f, "%d ",k);
    for l=1:columns(X)
      if ( strcmpi(tt,"float") )
	fprintf(f, "%f ",X(k,l));
      else
	fprintf(f, "%d ",X(k,l));
      endif
    endfor
    fprintf(f, "\n");
  endfor
  fprintf(f, ";\n");
endfunction

##############################################################################
## Solve the MIP problem using GLPK. 
##
## OUTPUT
##
## bid(r,i) = j iff user i requests Req(r,i) items of resource r from j
##		(if bid(r,i) == 0, then user i does not request any resource
##		from any seller)
function bid = solve_with_glpk( basename, Req, Off, M )
  assert( nargin == 4 );
  R = rows(Req);
  N = columns(Req);
  modname = "ecosystem.mod";
  datname = strcat(basename, ".dat"); # data file
  outname = strcat(basename, ".out"); # solution file

  ## check whether the otuput file already exists
  if ( 2 != exist(outname, "file") )
    
    f = fopen(datname,"w");
    fprintf(f, "data;\n");
    fprintf(f, "param N := %d;\n", N);
    fprintf(f, "param R := %d;\n", R);
    dump_glpk_matrix(f, "Req", Req);
    dump_glpk_matrix(f, "Off", Off);
    dump_glpk_matrix(f, "M", M);
    fprintf(f, "end;\n");
    fclose(f);
    cmd = sprintf("glpsol -m %s -d %s -y %s --tmlim 60 > /dev/null", modname, \
		  datname, outname);
    if ( system(cmd) != 0 )
      error("Execution of command \"%s\" failed.", cmd);
    endif

  endif

  load(outname,"bid");
endfunction

##############################################################################
## Make a set of simulation runs with given values of R, N and 
##
## INPUTS:
##
## basename base name of the output data files
## N	number of users
## R	number of resource types
## fb	fraction of buyers (0 < fb < 1)
## budget budget(i) is the initial budget of node i
## density average link density (fraction of nonzero elements in the adjacency matrix)
## T	number of time steps
##
## OUTPUTS:
##
## n_buyers(t)	number of buyers at step t
## n_opt_buyers(t) maximum number of buyers at step t (computed with GLPK)
## bb(t,i)	budget of user i at the end of step t
## max_iter	max number of iterations of the auction
## mean_price   (scalar) average unitary resource price for this run
## fairness_auction variance of the ratio of total amount of resources provided to i and requested by i; if i did not request anything, set to 0
## fairness_glpk
function [n_buyers n_opt_buyers bb max_iter mean_price fairness_auction fairness_glpk] = do_simulation_run( basename, N, R, fb, budget, density, T )

  assert( nargin == 7 );
  assert( ischar(basename) );
  assert( isscalar(N) );
  assert( isscalar(R) );
  assert( isscalar(fb) );
  assert( isvector(budget) );
  assert( N == length(budget) );
  assert( isscalar(density) );
  assert( isscalar(T) );

  resultfile = [basename ".octave"]; 
  if ( 2 == exist(resultfile) )
    load(resultfile);
    rand("state", rand_state);
    return;
  endif
    
  n_buyers = n_opt_buyers = zeros(1,T);
  mean_prices = zeros(1,T); # mean_prices(t) is the mean price at time t
  fairness_auction = fairness_glpk = bb = zeros(T,N);
  max_iter = 0;

  ## tot_requested(i) = total amount of resources requested by i
  ## tot_obtained_*(i) = total amount of resources obtained by i
  tot_requested = tot_obtained_glpk = tot_obtained_auction = zeros(1,N); 

  for t=1:T    
    
    M = make_overlay(N,density);
    
    [Off Req] = setup_round( N, R, fb );

    tot_requested += sum(Req,1);
    
    ## Compute optimal solution using GLPK
    datafile = sprintf("%s_T%02d",basename,t);
    tic();
    opt_bid = solve_with_glpk( datafile, Req, Off, M );
    time_glpk = toc();
    glpk_n_buyers = sum(any(opt_bid>0,1)); # maximum number of buyers
    tot_obtained_glpk(any(opt_bid>0,1)) += sum( Req(opt_bid>0), 1 );

    ## Compute solution using auctions
    SP = ones(R,N); # SP(r,j) is the unitary selling price for resource r sold by user j
    
    ## RP(i) is the reserve price of user i (i.e., the maximum amount user i will bid)
    RP = unifrnd(1.5,2,1,N) .* sum( Req, 1 );
    RP = min(RP, budget );
    
    nbuyers0 = sum(any(Req>0,1)); # number of buyers
    
    ## Main loop
    iter = 0;
    tmpReq = Req;
    tic();

    do
      bid = make_bids(tmpReq, RP, Off, SP, M);
      ## remove requests from users which made no bid
      tmpReq(bid==0) = 0;
      
      demand = compute_demands(bid,tmpReq);
      
      excess = demand>Off; # excess(r,j) = 1 iff demand(r,j) for type r resource at seller j is greater than the amount offered Off(r,j)
      deltaP = unifrnd(0.9,1.1,size(demand));
      SP(excess) += deltaP(excess); # increase price for excess resources
      
      iter++;
    until all(all(excess==0));
    time_auction = toc();

    max_iter = max( max_iter, iter );
    
    iscorrect( bid, Req, Off, M );

    tot_obtained_auction(any(bid>0,1)) += sum( Req(bid>0), 1 );    

    [expenses incomes] = compute_expense( bid, Req, SP );
    
    nbuyers1 = sum(any(bid>0,1)); # number of buyers

    if ( nbuyers1 > 0 ) # if there have been buyers
      mean_prices(t) = sum(expenses) / sum(sum(Req(bid>0))); # attention: not all requests are satisfied, so we must use Req(bid>0) to only select those requests which indeed have been satisfied
    else
      mean_prices(t) = 0;
    endif

    assert( glpk_n_buyers >= nbuyers1 );
    printf("%3d/%3d/%3d | Tokens:%4d | Av. Price:%5.2f | iter:%3d | T(GLPK):%3d | T(Auction):%3d\n", ...
	   nbuyers0, ...
	   nbuyers1, ...
	   glpk_n_buyers, ...
	   sum(expenses), ...
	   mean_prices(t), ...
	   iter, ...
	   time_glpk, ...
	   time_auction );
      
    budget = budget + incomes - expenses; # update budget    
    n_buyers(t) = nbuyers1;
    n_opt_buyers(t) = glpk_n_buyers;
    bb(t,:) = budget;
  endfor

  mean_price = mean(mean_prices);

  ratio_glpk = tot_obtained_glpk ./ tot_requested;
  ratio_auction = tot_obtained_auction ./ tot_requested;

  fairness_glpk = var(ratio_glpk(!isnan(ratio_glpk)));
  fairness_auction = var(ratio_auction(!isnan(ratio_auction)));

  rand_state = rand("state");
  save(resultfile, "n_buyers","n_opt_buyers","bb","rand_state", "max_iter", "mean_price", "fairness_auction", "fairness_glpk");
endfunction

##############################################################################
## Compute average (x) and half length confidence interval (hl)
## for "samples" at confidence level "confidence" (e.g., 0.95)
## source: http://www.omnetpp.org/listarchive/msg00229.php
function [x hl] = estimate_mean( samples, confidence )
  if ( nargin < 2 )
    confidence = .9;
  endif

  alpha = 1-confidence; # if confidence=0.95, alpha=0.05
  ## retrieve sample size
  n = length(samples);
  ## calculate the mean of the samples
  x = mean(samples);
  ## calculate the half length confidence interval
  ## hl = tinv((0.5+(confidence/200)), n-1)*std(samples)/sqrt(n);
  hl = tinv(1-alpha/2, n-1)*std(samples)/sqrt(n);
endfunction

##############################################################################
## Perform multiple independent simulation runs. _hw are confidence
## intervals half-width. For the budget, no confidence intervals are
## computed (as they are not useful anyway).
##
## OUTPUT
##
## result.n_buyers
## result.n_buyers_hw
## result.n_opt_buyers
## result.n_opt_buyers_hw
## result.n_iter		(average) max number of iterations
## result.bb
## result.tot_buyers		Total number of matches
## result.tot_buyers_hw		Conf. interval half width
## result.tot_opt_buyers	Optimal number of matches
## result.tot_opt_buyers_hw	Conf. interval half width
## result.mean_price		Average cost of a single resource
function result = do_simulation( basename, N, R, fb, budget, density, T, nreps )

  assert( nargin == 8 );

  ## Perform runs and accumulate results
  bb = zeros(T,N);
  n_buyers_acc = n_opt_buyers_acc = zeros(nreps,T);
  mean_price_acc = n_iter_acc = zeros(1,nreps);
  fairness_auction_acc = fairness_glpk_acc = zeros(nreps,N);

  for nrep=1:nreps
    printf("***\n*** Replication %d/%d (N=%d, R=%d, density=%.2f)\n***\n", nrep, nreps, N, R, density);
    basename_rep = sprintf("%s_rep%02d",basename,nrep);
    [n_buyers_rep n_opt_buyers_rep bb_rep max_iter_rep mean_price_rep] = do_simulation_run( basename_rep, N, R, fb, budget, density, T );
    n_iter_acc(nrep) = max_iter_rep;
    n_buyers_acc(nrep,:) = n_buyers_rep;
    n_opt_buyers_acc(nrep,:) = n_opt_buyers_rep;
    mean_price_acc(nrep) = mean_price_rep;
    bb += bb_rep;
  endfor
  
  result.bb = bb / nreps; ## do not compute confidence intervals
  result.n_iter = mean(n_iter_acc);
  result.mean_price = mean(mean_price_acc);
  ## Compute confidence intervals from results
  result.n_buyers = result.n_buyers_hw = ...
      result.n_opt_buyers = result.n_opt_buyers_hw = zeros(1,T);
  for t=1:T
    [result.n_buyers(t) result.n_buyers_hw(t)] = estimate_mean( n_buyers_acc(:,t) );
    [result.n_opt_buyers(t) result.n_opt_buyers_hw(t)] = estimate_mean( n_opt_buyers_acc(:,t) );
  endfor
  [result.tot_buyers result.tot_buyers_hw] = ...
      estimate_mean( sum(n_buyers_acc,1) );
  [result.tot_opt_buyers result.tot_opt_buyers_hw] = ...
      estimate_mean( sum(n_opt_buyers_acc,1) );
endfunction

##############################################################################
## Plot error bars over histograms
function h = barerror( barvalues, errorvalues, leg )
  hold on;
  bars = bar( barvalues );
  if ( nargin == 3 )
    legend( leg );
  endif
  h = gca();
  numgroups = rows(barvalues);
  numbars = columns(barvalues);
  markers = "ox.";
  for i=1:numbars
    x = get(get(bars(i),"children"),"xdata");
    d = mean(x([1 3],:));
    for j=1:numgroups
      tmp = errorbar(d(j), barvalues(j,i), errorvalues(j,i));
      set(tmp,"color",[0 0 0]); # error bars are black
      set(tmp,"linewidth",2); # set width of error bars
      set(tmp,"marker",markers(i));
    endfor
  endfor
  hold off;
endfunction

##############################################################################
##############################################################################
##############################################################################

#{

N = 50;
R = 7;
T = 50; # number of time steps
nreps = 20; # number of replications
density = .3;
fb = .2; # fraction of buyers
budget = 100*ones(1,N); # budget(i) is the total budget of user i
[n_buyers n_opt_buyers bb max_iter mean_price fairness_auction fairness_glpk] = do_simulation_run("test_variance", N, R, fb, budget, density, T );

fairness_auction
fairness_glpk

return;

#}

##
## Initialize rng
##
rand("state", [1 1 2 3 5 8 13 21 34]);

## 
## Global parameters
##
T = 10; # number of time steps
nreps = 20; # number of replications
density = .3;

## Setup plot area
set(gcf,"paperorientation","landscape");
papersize=[5 4] * 0.8; 
margin=[0 0];
set(gcf,"papersize",papersize);
set(gcf,"paperposition", [margin papersize-margin*2]);
colormap([.8 .8 .8; .55 .55 .55]); # gray bars

## Setup arrays for total number of matches
fb = .2; # fraction of buyers
allR = [3 5 7];
allN = [10 20 50];
tot_buyers = tot_buyers_hw = ...
    tot_opt_buyers = tot_opt_buyers_hw = ...
    n_iter = ...
    zeros( 1, length(allR)*length(allN) );
lab = cell(1,length(allR)*length(allN)); # labels of the histogram
idx = 1;

## setup output data file
f = fopen("results.tex","w");

for R=allR
  for N=allN

    budget = 100*ones(1,N); # budget(i) is the total budget of user i
    testname = sprintf("test1_N%03d_R%02d",N,R);
    result = do_simulation(testname, N, R, fb, budget, density, T, nreps );

#{    
    ## Plot average number of buyers vs maximum number of buyers at each step
    clf;
    h = barerror([result.n_buyers' result.n_opt_buyers'], ...
		 [result.n_buyers_hw' result.n_opt_buyers_hw'] );
    xlabel("Time step");
    ylabel("Matched Requests");
    ax = axis();
    ax(2) = T+1;
    ax(3) = 0;
    ax(4) = max( max( [result.n_buyers; result.n_opt_buyers] ) ) * 1.2;
    axis(ax);
    legend("Auction", "Optimal");
    title(sprintf("%d users, %d resources",N,R));
    histname = sprintf("hist_N%03d_R%02d",N,R);
    print([histname ".eps"], "-depsc2", "-tight");
#}
    tot_buyers(idx) = result.tot_buyers;
    tot_buyers_hw(idx) = result.tot_buyers_hw;
    tot_opt_buyers(idx) = result.tot_opt_buyers;
    tot_opt_buyers_hw(idx) = result.tot_opt_buyers_hw;
    n_iter(idx) = result.n_iter;
    lab{idx} = sprintf("%d / %d",N,R);
    idx++;

    fprintf(f,"%d & %d & $%.2f \\pm %.2f$ & $%.2f \\pm %.2f$ & %.2f \\\\\n", N, R, ...
	    result.tot_buyers, result.tot_buyers_hw, ...
	    result.tot_opt_buyers, result.tot_opt_buyers_hw, ...
	    result.tot_buyers / result.tot_opt_buyers);
    fflush(f);
  endfor
  if ( R < max(allR) )
    fprintf(f, "\\addlinespace\n");
  endif
endfor

fclose(f);

##############################################################################
## Plot distribution of budget at step T/2 and T
N = 50; 
R = 7;
fb = .2;
density = .3;
budget = 100*ones(1,N);
testname = sprintf("test1_N%03d_R%02d",N,R);
result = do_simulation(testname, N, R, fb, budget, density, T, nreps );

[YY1 XX1] = hist(result.bb(fix(T/2),:));
[YY2 XX2] = hist(result.bb(T,:));

xmax = max( [XX1 XX2] )*1.2;
ymax = max( [YY1 YY2] )*1.2;

clf;
subplot(2,1,1);
hist(result.bb(fix(T/2),:),5:10:xmax);
ylabel("Number of users");
title(sprintf("After %d steps",fix(T/2)));
ax = axis();
ax(1) = 0;
ax(2) = xmax;
ax(3) = 0;
ax(4) = ymax;
axis(ax);

subplot(2,1,2);
hist(result.bb(T,:),5:10:xmax);
xlabel("Budget");
ylabel("Number of users");
title(sprintf("After %d steps",T));
ax = axis();
ax(1) = 0;
ax(2) = xmax;
ax(3) = 0;
ax(4) = ymax;
axis(ax);
print( sprintf("hist_N%03d_R%02d_bb.eps", N, R), "-depsc2", "-tight");


##############################################################################
## PLot histogram showing the total number of matches
clf;
set(gcf,"paperorientation","landscape");
papersize=[5 2.5] * 0.8; 
margin=[0 0];
set(gcf,"papersize",papersize);
set(gcf,"paperposition", [margin papersize-margin*2]);
bar([tot_buyers' tot_opt_buyers'])
xlabel("Users (N) / Resource Types (R)");
ylabel("Number of matches");
h = gca();
set(h,"xtick",1:length(allR)*length(allN));
set(h,"xticklabel",lab);
legend({"Auction","Optimal"}, "location", "northeast");
print("hist_tot.eps","-depsc2","-tight");


##############################################################################
## What happens for various network densities?
N = 50; 
R = 7;
fb = .2;
budget = 100*ones(1,N);
densities=[0.2 0.4 0.6 0.8];
tot_buyers = tot_opt_buyers = mean_price = n_iter = zeros(size(densities));
lab = cell(1,length(densities));

f = fopen("results_density.tex","w");

for idx=1:length(densities)
  density=densities(idx);
  testname = sprintf("test_density_%f",density);
  result = do_simulation(testname, N, R, fb, budget, density, T, nreps );
  tot_buyers(idx) = result.tot_buyers;
  tot_opt_buyers(idx) = result.tot_opt_buyers;
  mean_price(idx) = result.mean_price;
  n_iter(idx) = result.n_iter;

  fprintf(f, "%.2f & $%.2f \\pm %.2f$ & $%.2f \\pm %.2f$ & %.2f & %d\\\\\n", ...
	  density, ...
	  result.tot_buyers, result.tot_buyers_hw, ...
	  result.tot_opt_buyers, result.tot_opt_buyers_hw,...
	  result.mean_price, ...
	  result.n_iter );
  fflush(f);

  lab{idx} = sprintf("%.2f", density);
endfor

fclose(f);

clf;
bar([tot_buyers' tot_opt_buyers'] );
xlabel("Density of connections (\\rho)");
ylabel("Number of matches");
h = gca();
set(h,"xtick",1:length(densities));
set(h,"xticklabel",lab);
legend({"Auction","Optimal"}, "location", "northwest");
print("hist_densities.eps","-depsc2","-tight");

clf;
subplot(2,1,1);
plot(densities,mean_price,"-o","linewidth",2);
h = gca();
set(h,"xtick",densities);
ylabel("Unitary res. prices");
ax = axis();
ax(3) = 0;
axis(ax);

subplot(2,1,2);
plot(densities,n_iter,"-o","linewidth",2);
h = gca();
set(h,"xtick",densities);
xlabel("Density of connections (\\rho)");
ylabel("Number of rounds");
ax = axis();
ax(3) = 0;
axis(ax);

print("plot_prices_iterations.eps","-depsc2", "-tight");
