## This script implements the Linear Integer Programming model used to
## compute the optimal allocation described in in M. Marzolla,
## S. Ferretti, G. D'Angelo.  "Auction-Based Resource Allocation in
## Digital Ecosystems", Proc. 6th International Conference on MOBILe Wireless
## MiddleWARE, Operating Systems, and Applications (MobilWare 2013),
## pp. 20-27, nov 11-12 2013, Bologna, Italy, DOI http://dx.doi.org/10.1109/Mobilware.2013.16
##
## This script has been written in july-august 2011 by Moreno Marzolla
## <marzolla (at) cs.unibo.it> and released by the author in the Public
## Domain.
##
## This script is called by ecosystem.m

param N > 0, integer; # number of users

param R > 0, integer; # Number of resource types

set Users := 1..N; # Set of users

set Resources := 1..R; # Set of resources

param Req{Resources, Users} >= 0; # Req[r,u] is the amount of resource r requested by user u

param Off{Resources, Users} >= 0; # Off[r,u] is the amount of resource r offered by user u

param M{Users, Users}, binary; # M[i,j] = 1 iff user i can interact with user j; matrix M must be symmetric

var X{Users, Resources, Users}, binary; # X[i,r,j] = 1 iff user i gets resource r from user j

## Optimization problem: maximize the number of users with requests
## satisfied. We should divide the objective function by R, but since
## this is a constant, we don't care.'
maximize num_sat_users: sum{i in Users, r in Resources, j in Users} X[i,r,j];

## The total amount of resource provided by user j must be less than
## or equal to the amount obtained by other users
s.t. max_capacity {j in Users, r in Resources}: sum{i in Users} Req[r,i]*X[i,r,j] <= Off[r,j];

## Each resource requested must be provided by at most one user (i.e.,
## it is not possible to get partial amounts of resource from
## different users).
s.t. single_seller {r in Resources, i in Users}: sum{j in Users} X[i,r,j] <= 1;

## Each user must get all the resources he requests, or none at all
s.t. everything_or_nothing {i in Users, r in Resources}: sum{j in Users} X[i,r,j] = sum{j in Users} X[i,1,j];

## User i can buy from j iff i is connected to j
s.t. connected {i in Users, r in Resources, j in Users}: X[i,r,j] <= M[i,j];

## If Req[r,i] == 0 (user i does not request resource r), then we set X[i,r,j] = 0 for all j
s.t. no_res_requested {i in Users, r in Resources, j in Users : Req[r,i]==0}: X[i,r,j] = 0;

solve;

## Display the solution in a format that Octave can read as a text data file

printf "# Created by GLPK\n";
printf "# name: bid\n";
printf "# type: matrix\n";
printf "# rows: %d\n", R;
printf "# columns: %d\n", N;
for {r in Resources} {
for {i in Users} {
## printf "%d ", sum{j in Users} (if Req[r,i]>0 then 1 else 0)*j*X[i,r,j];
printf "%d ", sum{j in Users} j*X[i,r,j];
}
printf("\n");
}

data;

param N := 5;

param R := 3;

param Off :  1  2  3  4  5 :=
	1    5  7  7 10  5
	2   10  5  3 12 10
        3    7  5  3  8  5 ;


param Req :  1  2  3  4  5 :=
	1    0  4  0  0  5
        2    0  7  0  0  3
        3    0  1  0  0  2 ;

param M : 1 2 3 4 5 :=
	1 0 1 0 0 0
	2 1 0 1 1 0
	3 0 1 0 0 1
	4 0 1 0 0 0
	5 0 0 1 0 0 ;

end;

